@@ -2,8 +2,17 @@
 {-# LANGUAGE DeriveDataTypeable  #-}
 {-# LANGUAGE LambdaCase  #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE FlexibleContexts #-}
 module Tests.Control.Retry
     ( tests
+
+    -- * Used to test UnliftIO versions of the same functions
+    , recoveringTestsWith
+    , maskingStateTestsWith
+    , quadraticDelayTestsWith
+    , recoveringTest
+    , testHandlers
+    , testHandlersDynamic
     ) where
 
 -------------------------------------------------------------------------------
@@ -11,9 +20,11 @@ import           Control.Applicative
 import           Control.Concurrent
 import           Control.Concurrent.STM      as STM
 import qualified Control.Exception           as EX
+import           Control.Monad               as M ( forM_ )
 import           Control.Monad.Catch
+import           Control.Monad.Except
 import           Control.Monad.Identity
-import           Control.Monad.IO.Class
+import           Control.Monad.IO.Class      as MIO
 import           Control.Monad.Writer.Strict
 import           Data.Either
 import           Data.IORef
@@ -48,17 +59,25 @@ tests = testGroup "Control.Retry"
   , limitRetriesByCumulativeDelayTests
   , overridingDelayTests
   , resumableTests
+  , retryOnErrorTests
   ]
 
 
 -------------------------------------------------------------------------------
 recoveringTests :: TestTree
-recoveringTests = testGroup "recovering"
+recoveringTests = recoveringTestsWith recovering
+
+
+recoveringTestsWith
+    :: Monad m
+    => (RetryPolicyM m -> [RetryStatus -> Handler IO Bool] -> (a -> IO ()) -> IO ())
+    -> TestTree
+recoveringTestsWith recovering' = testGroup "recovering"
   [ testProperty "recovering test without quadratic retry delay" $ property $ do
       startTime <- liftIO getCurrentTime
       timeout <- forAll (Gen.int (Range.linear 0 15))
       retries <- forAll (Gen.int (Range.linear 0 50))
-      res <- liftIO $ try $ recovering
+      res <- liftIO $ try $ recovering'
         (constantDelay timeout <> limitRetries retries)
         testHandlers
         (const $ throwM (userError "booo"))
@@ -85,7 +104,7 @@ recoveringTests = testGroup "recovering"
 
       , testCase "recovers from custom exceptions" $ do
           f <- mkFailN Custom1 2
-          res <- try $ recovering
+          res <- try $ recovering'
             (constantDelay 5000 <> limitRetries 3)
             [const $ Handler $ \ Custom1 -> return shouldRetry]
             f
@@ -93,7 +112,7 @@ recoveringTests = testGroup "recovering"
 
       , testCase "fails beyond policy using custom exceptions" $ do
           f <- mkFailN Custom1 3
-          res <- try $ recovering
+          res <- try $ recovering'
             (constantDelay 5000 <> limitRetries 2)
             [const $ Handler $ \ Custom1 -> return shouldRetry]
             f
@@ -108,7 +127,7 @@ recoveringTests = testGroup "recovering"
 
       , testCase "does not recover from unhandled exceptions" $ do
           f <- mkFailN Custom2 2
-          res <- try $ recovering
+          res <- try $ recovering'
             (constantDelay 5000 <> limitRetries 5)
             [const $ Handler $ \ Custom1 -> return shouldRetry]
             f
@@ -117,7 +136,7 @@ recoveringTests = testGroup "recovering"
 
       , testCase "recovers in presence of multiple handlers" $ do
           f <- mkFailN Custom2 2
-          res <- try $ recovering
+          res <- try $ recovering'
             (constantDelay 5000 <> limitRetries 5)
             [ const $ Handler $ \ Custom1 -> return shouldRetry
             , const $ Handler $ \ Custom2 -> return shouldRetry ]
@@ -127,7 +146,7 @@ recoveringTests = testGroup "recovering"
 
       , testCase "general exceptions catch specific ones" $ do
           f <- mkFailN Custom2 2
-          res <- try $ recovering
+          res <- try $ recovering'
             (constantDelay 5000 <> limitRetries 5)
             [ const $ Handler $ \ (_::SomeException) -> return shouldRetry ]
             f
@@ -136,7 +155,7 @@ recoveringTests = testGroup "recovering"
 
       , testCase "(redundant) even general catchers don't go beyond policy" $ do
           f <- mkFailN Custom2 3
-          res <- try $ recovering
+          res <- try $ recovering'
             (constantDelay 5000 <> limitRetries 2)
             [ const $ Handler $ \ (_::SomeException) -> return shouldRetry ]
             f
@@ -146,7 +165,7 @@ recoveringTests = testGroup "recovering"
       , testCase "rethrows in presence of failed exception casts" $ do
           f <- mkFailN Custom2 3
           final <- try $ do
-            res <- try $ recovering
+            res <- try $ recovering'
               (constantDelay 5000 <> limitRetries 2)
               [ const $ Handler $ \ (_::SomeException) -> return shouldRetry ]
               f
@@ -227,10 +246,17 @@ policyTransformersTests = testGroup "policy transformers"
 
 -------------------------------------------------------------------------------
 maskingStateTests :: TestTree
-maskingStateTests = testGroup "masking state"
+maskingStateTests = maskingStateTestsWith recovering
+
+
+maskingStateTestsWith
+    :: Monad m
+    => (RetryPolicyM m -> [RetryStatus -> Handler IO Bool] -> (a -> IO b) -> IO ())
+    -> TestTree
+maskingStateTestsWith recovering' = testGroup "masking state"
   [ testCase "shouldn't change masking state in a recovered action" $ do
       maskingState <- EX.getMaskingState
-      final <- try $ recovering retryPolicyDefault testHandlers $ const $ do
+      final <- try $ recovering' retryPolicyDefault testHandlers $ const $ do
         maskingState' <- EX.getMaskingState
         maskingState' @?= maskingState
         fail "Retrying..."
@@ -245,7 +271,7 @@ maskingStateTests = testGroup "masking state"
                 maskingState @?= EX.MaskedInterruptible
                 return shouldRetry
             ]
-      final <- try $ recovering retryPolicyDefault checkMaskingStateHandlers $ const $ fail "Retrying..."
+      final <- try $ recovering' retryPolicyDefault checkMaskingStateHandlers $ const $ fail "Retrying..."
       assertBool
         ("Expected EX.IOException but didn't get one")
         (isLeft (final :: Either EX.IOException ()))
@@ -260,8 +286,8 @@ capDelayTests = testGroup "capDelay"
       cap <- forAll (Gen.int (Range.linear 1 maxBound))
       let policy = capDelay cap (limitRetries retries)
       let delays = runIdentity (simulatePolicy (retries + 1) policy)
-      let Just lastDelay = lookup (retries - 1) delays
-      let Just gaveUp = lookup retries delays
+      let lastDelay = fromMaybe (error "impossible: empty delays") (lookup (retries - 1) delays)
+      let gaveUp = fromMaybe (error "impossible: empty delays") (lookup retries delays)
       let noDelay = 0
       lastDelay === Just noDelay
       gaveUp === Nothing
@@ -301,12 +327,19 @@ limitRetriesByCumulativeDelayTests = testGroup "limitRetriesByCumulativeDelay"
 
 -------------------------------------------------------------------------------
 quadraticDelayTests :: TestTree
-quadraticDelayTests = testGroup "quadratic delay"
+quadraticDelayTests = quadraticDelayTestsWith recovering
+
+
+quadraticDelayTestsWith
+    :: Monad m
+    => (RetryPolicyM m -> [RetryStatus -> Handler IO Bool] -> (a -> IO b) -> IO ())
+    -> TestTree
+quadraticDelayTestsWith recovering' = testGroup "quadratic delay"
   [ testProperty "recovering test with quadratic retry delay" $ property $ do
       startTime <- liftIO getCurrentTime
       timeout <- forAll (Gen.int (Range.linear 0 15))
       retries <- forAll (Gen.int (Range.linear 0 8))
-      res <- liftIO $ try $ recovering
+      res <- liftIO $ try $ recovering'
         (exponentialBackoff timeout <> limitRetries retries)
         [const $ Handler (\(_::SomeException) -> return True)]
         (const $ throwM (userError "booo"))
@@ -356,7 +389,7 @@ overridingDelayTests = testGroup "overriding delay"
         (handler delays)
         (action delays)
       let expectedDelays = map microsToNominalDiffTime delays
-      forM_ (zip (diffTimes measuredTimestamps) expectedDelays) $
+      M.forM_ (zip (diffTimes measuredTimestamps) expectedDelays) $
         \(actual, expected) -> diff actual (>=) expected
 
 
@@ -386,69 +419,85 @@ resumableTests = testGroup "resumable"
             ()
       ]
   ]
-  where
-    retryingTest
-      :: (RetryStatus -> RetryPolicyM IO -> p -> (RetryStatus -> IO ()) -> IO ())
-      -> p
-      -> IO ()
-    retryingTest resumableOp isRetryNeeded = do
-      counterRef <- newIORef (0 :: Int)
 
-      let go policy status = do
-            atomicWriteIORef counterRef 0
-            resumableOp
-              status
-              policy
-              isRetryNeeded
-              (const $ atomicModifyIORef' counterRef $ \n -> (1 + n, ()))
+retryingTest
+  :: (RetryStatus -> RetryPolicyM IO -> p -> (RetryStatus -> IO ()) -> IO ())
+  -> p
+  -> IO ()
+retryingTest resumableOp isRetryNeeded = do
+  counterRef <- newIORef (0 :: Int)
 
+  let go policy status = do
+        atomicWriteIORef counterRef 0
+        resumableOp
+          status
+          policy
+          isRetryNeeded
+          (const $ atomicModifyIORef' counterRef $ \n -> (1 + n, ()))
+
+  let policy = limitRetries 2
+  let nextStatus = nextStatusUsingPolicy policy
+
+  go policy defaultRetryStatus
+  (3 @=?) =<< readIORef counterRef
+
+  go policy =<< nextStatus defaultRetryStatus
+  (2 @=?) =<< readIORef counterRef
+
+  go policy =<< nextStatus =<< nextStatus defaultRetryStatus
+  (1 @=?) =<< readIORef counterRef
+
+recoveringTest
+  :: (RetryStatus -> RetryPolicyM IO -> handlers -> (RetryStatus -> IO ()) -> IO ())
+  -> handlers
+  -> IO ()
+recoveringTest resumableOp handlers = do
+  counterRef <- newIORef (0 :: Int)
+
+  let go policy status = do
+        action <- do
+          mkFailUntilIO
+            (\_ -> atomicModifyIORef' counterRef $ \n -> (1 + n, False))
+            Custom1
+        try $ resumableOp status policy handlers action
+
+  let policy = limitRetries 2
+  let nextStatus = nextStatusUsingPolicy policy
+
+  do
+    atomicWriteIORef counterRef 0
+    res <- go policy defaultRetryStatus
+    res @?= Left Custom1
+    (3 @=?) =<< readIORef counterRef
+
+  do
+    atomicWriteIORef counterRef 0
+    res <- go policy =<< nextStatus defaultRetryStatus
+    res @?= Left Custom1
+    (2 @=?) =<< readIORef counterRef
+
+  do
+    atomicWriteIORef counterRef 0
+    res <- go policy =<< nextStatus =<< nextStatus defaultRetryStatus
+    res @?= Left Custom1
+    (1 @=?) =<< readIORef counterRef
+
+
+-------------------------------------------------------------------------------
+retryOnErrorTests :: TestTree
+retryOnErrorTests = testGroup "retryOnError"
+  [ testCase "passes in the error type" $ do
+      errCalls <- newTVarIO []
       let policy = limitRetries 2
-      let nextStatus = nextStatusUsingPolicy policy
-
-      go policy defaultRetryStatus
-      (3 @=?) =<< readIORef counterRef
-
-      go policy =<< nextStatus defaultRetryStatus
-      (2 @=?) =<< readIORef counterRef
-
-      go policy =<< nextStatus =<< nextStatus defaultRetryStatus
-      (1 @=?) =<< readIORef counterRef
-
-    recoveringTest
-      :: (RetryStatus -> RetryPolicyM IO -> handlers -> (RetryStatus -> IO ()) -> IO ())
-      -> handlers
-      -> IO ()
-    recoveringTest resumableOp handlers = do
-      counterRef <- newIORef (0 :: Int)
-
-      let go policy status = do
-            action <- do
-              mkFailUntilIO
-                (\_ -> atomicModifyIORef' counterRef $ \n -> (1 + n, False))
-                Custom1
-            try $ resumableOp status policy handlers action
-
-      let policy = limitRetries 2
-      let nextStatus = nextStatusUsingPolicy policy
-
-      do
-        atomicWriteIORef counterRef 0
-        res <- go policy defaultRetryStatus
-        res @?= Left Custom1
-        (3 @=?) =<< readIORef counterRef
-
-      do
-        atomicWriteIORef counterRef 0
-        res <- go policy =<< nextStatus defaultRetryStatus
-        res @?= Left Custom1
-        (2 @=?) =<< readIORef counterRef
-
-      do
-        atomicWriteIORef counterRef 0
-        res <- go policy =<< nextStatus =<< nextStatus defaultRetryStatus
-        res @?= Left Custom1
-        (1 @=?) =<< readIORef counterRef
-
+      let shouldWeRetry _retryStat e = do
+            liftIO (atomically (modifyTVar' errCalls (++ [e])))
+            return True
+      let action rs = (throwError ("boom" ++ show (rsIterNumber rs)))
+      res <- runExceptT (retryOnError policy shouldWeRetry action)
+      res @?= (Left "boom2" :: Either String ())
+      calls <- atomically (readTVar errCalls)
+      calls @?= ["boom0", "boom1", "boom2"]
+  ]
 
 -------------------------------------------------------------------------------
 nextStatusUsingPolicy :: RetryPolicyM IO -> RetryStatus -> IO RetryStatus
@@ -513,7 +562,7 @@ genRetryStatus = do
 -------------------------------------------------------------------------------
 -- | Generate an arbitrary 'RetryPolicy' without any limits applied.
 genPolicyNoLimit
-    :: forall mg mr. (MonadGen mg, MonadIO mr)
+    :: forall mg mr. (MonadGen mg, MIO.MonadIO mr)
     => Range Int
     -> mg (RetryPolicyM mr)
 genPolicyNoLimit durationRange =
